import unittest
from unittest.mock import patch

import numpy as np

from counterplots.utils.plots import make_greedy_plot, make_countershapley_plot, make_constellation_plot


class TestMakeGreedyPlot(unittest.TestCase):
    @patch('counterplots.utils.plots.plt')
    @patch('counterplots.utils.plots.mpl')
    def test_make_greedy_plot(self, mock_mpl, mock_plt):
        factual_score = 0.7
        features_data = [
            {'score': 0.4, 'name': 'Feature 1', 'factual': 'A', 'counterfactual': 'B'},
            {'score': 0.6, 'name': 'Feature 2', 'factual': 'X', 'counterfactual': 'Y'}
        ]
        class_names = {0: 'Class A', 1: 'Class B'}
        save_path = 'plot.png'

        make_greedy_plot(factual_score, features_data, class_names, save_path)


class TestMakeCounterShapleyPlot(unittest.TestCase):

    @patch('counterplots.utils.plots.plt')
    @patch('counterplots.utils.plots.mpl')
    def test_make_countershapley_plot(self, mock_mpl, mock_plt):
        factual_score = 0.0
        features_data = [{'x': 62.0,
                          'score': 0.4,
                          'name': 'Feature 1',
                          'factual': -4.0,
                          'counterfactual': -1.0},
                         {'x': 38.0,
                          'score': 0.6,
                          'name': 'Feature 2',
                          'factual': 1.0,
                          'counterfactual': -1.0}]
        class_names = {0: 'Class A', 1: 'Class B'}
        save_path = 'plot.png'

        make_countershapley_plot(factual_score, features_data, class_names, save_path)


class TestMakeConstellationPlot(unittest.TestCase):

    @patch('counterplots.utils.plots.plt')
    @patch('counterplots.utils.plots.mpl')
    def test_make_constellation_plot(self, mock_mpl, mock_plt):
        factual_score = 0.0
        single_points_chart = np.array([[0., 0.00519673],
                                        [1., 0.06478132],
                                        [2., 0.004789]])
        text_features = ['Feature 6 (-3.19627➜-0.3652)',
                         'Feature 8 (2.0093➜-3.15686)',
                         'Feature 9 (3.1568➜-2.41393)']
        mulitple_points_chart = np.array(
            [[np.array([0, 1]), 0.263131977249517],
             [np.array([0, 2]), 0.024619989978766642],
             [np.array([1, 2]), 0.24759475995893304]], dtype='object')
        mulitple_points_chart_y = [0.5, 1.0, 1.5]
        single_points = np.array([
            [6.00000000e+00, 5.19673099e-03],
            [8.00000000e+00, 6.47813170e-02],
            [9.00000000e+00, 4.78899690e-03]])
        class_names = {0: 'Class 1', 1: 'Class 0'}
        cf_score = 0.63
        point_to_pred = {0: 0.005196730993880675, 1: 0.06478131701518305, 2: 0.004788996904253137}
        save_path = 'plot.png'

        make_constellation_plot(factual_score, single_points_chart, text_features, mulitple_points_chart,
                                mulitple_points_chart_y, single_points, class_names, cf_score, point_to_pred,
                                save_path)



import numpy as np


# Verifications
def verify_shape(array):
    if len(np.array(array).shape) != 1:
        raise ValueError('Array must be 1-dimensional.')


def verify_factual_counterfactual_shape(factual, cf):
    for type, data in zip(['Factual', 'Counterfactual'], [factual, cf]):
        if len(data.shape) != 1:
            raise ValueError(f'{type} array must be 1-dimensional. Got {data.shape}.')


def verify_factual_cf(factual, cf):
    if np.array(factual).shape != np.array(cf).shape:
        raise ValueError('Factual and counterfactual must have the same shape.')


def verify_feature_names(factual, feature_names):
    if len(feature_names) != np.array(factual).shape[0]:
        raise ValueError('Feature names must have the same length as the number of features.')


def verify_model_prediction(model_pred, factual):
    try:
        model_pred(np.array([factual]))
    except Exception as e:
        raise ValueError('Model prediction function must be callable but got error: {}'.format(e))


def verify_binary_class(model_pred, factual):
    model_pred_shape = model_pred(np.array([factual])).shape
    if len(model_pred_shape) == 1:
        return
    elif len(model_pred_shape) == 2:
        if model_pred_shape[1] > 2:
            raise ValueError('Model must be binary.')
    else:
        raise ValueError('Model must be binary.')


def verify_class_names(class_names):
    if type(class_names) != dict or len(class_names) != 2:
        raise ValueError(
            'Class names must be a dictionary, where int(0) is the class for prediction 0 and '
            'int(1) is the class for prediction 1.')



import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.path import Path
import matplotlib.patches as patches


def make_greedy_plot(factual_score, features_data, class_names, save_path):
    fig = plt.figure(figsize=(7, 1 * len(features_data))) # noqa

    scatter_points = [f['score'] for f in features_data]
    scatter_points.insert(0, factual_score)

    plt.plot(scatter_points, range(len(scatter_points)),
             color='#c4c4c4', linestyle='dashed', zorder=0)
    for c_idx, point in enumerate(scatter_points):
        plt.scatter(
            [point],
            [c_idx],
            color='#ff0055D2' if point < 0.5 else '#008ae7D2',
            s=100)

    features_names = ['Factual']
    for feat_idx, f in enumerate(features_data):
        features_names.append(
            f'{f["name"]} ({f["factual"]}➜{f["counterfactual"]})')
    max_feat_names_length = max([len(feat_name)
                                for feat_name in features_names])

    # Plot a vertical line at the threshold score
    plt.axvline(x=0.5, color='#c20000', zorder=0, linewidth=0.5)

    # Plot a vertical line at the factual score
    plt.axvline(x=factual_score, color='#ff0055D2',
                linestyle='dashed', zorder=0)
    plt.text(factual_score - 0.088, len(features_names)*1.04 -
             0.9, 'Factual Score', color='#ff0055D2')

    # Plot a vertical line at the counterfactual score
    plt.axvline(x=scatter_points[-1],
                color='#008ae7D2', linestyle='dashed', zorder=0)
    plt.text(scatter_points[-1] - 0.15, len(features_names)*1.04 -
             0.9, 'Counterfactual Score', color='#008ae7D2')

    for feat_idx, feat_name in enumerate(features_names):
        plt.text(-0.02 * max_feat_names_length, feat_idx,
                 feat_name, color='#545454', fontsize=12)

    # Print binary class
    class_0_name = list(class_names.values())[0]
    class_1_name = list(class_names.values())[1]
    size_factual_class = mpl.textpath.TextPath(
        (0, 0), class_0_name).get_extents().width * 0.003
    plt.text(0.47 - size_factual_class - max([len(f) for f in features_names])*0.0001, len(features_names)
             - 0.75 + (len(features_names) + 1)*0.04,
             class_0_name, color='#ff0055D2', fontweight='bold')
    plt.text(0.49, len(features_names) - 0.75 + (len(features_names) + 1)*0.04, '➜',
             color='#c20000', fontweight='bold')
    plt.text(0.52, len(features_names) - 0.75 + (len(features_names) + 1)
             * 0.04, class_1_name, color='#008ae7D2', fontweight='bold')

    plt.gca().axes.get_yaxis().set_visible(False)
    plt.gca().spines['top'].set_visible(False)
    plt.gca().spines['left'].set_visible(False)
    plt.gca().spines['right'].set_visible(False)

    # Set limit of x axis from 0 to 1
    plt.xlim(-0.05, 1)

    if save_path:
        plt.savefig(save_path, bbox_inches='tight')


def make_countershapley_plot(factual_score, features_data, classes, save_path):
    fig = plt.figure(figsize=(10, 1.5))
    ax = fig.add_subplot(111)
    scale_x = 100
    scale_y = 100
    bar_y_height = scale_y / 2
    bar_y_height_sub = scale_y / 8.0
    fontsize = 10
    color_bar = '#48C975D2'
    cf_score = features_data[-1]['score']
    sum_countershapley = cf_score - factual_score

    # Beak must be always the 4%
    def create_bar(start_x, end_x, no_beak=False):

        verts = [
            (start_x, 0.),  # left, bottom
            (start_x, bar_y_height),  # left, top
            (end_x - 4, bar_y_height),  # right, top
            (end_x, bar_y_height),  # right, top
            (end_x, 0.),  # right, bottom
            (0., 0.),  # ignored
        ]

        codes = [
            Path.MOVETO,
            Path.LINETO,
            Path.LINETO,
            Path.LINETO if no_beak else Path.CURVE3,
            Path.LINETO,
            Path.CLOSEPOLY,
        ]

        return Path(verts, codes)

    prev_score = factual_score
    current_x = 0
    x_left_pos = []
    for _, feat_data in enumerate(features_data):
        x_left_pos.append(current_x)
        x_size = feat_data['x'] * 100 / scale_x + current_x

        # Plot text for feature name
        plt.text(
            current_x + (x_size - current_x) / 2 -
            len(feat_data['name']) * fontsize / 10 / 2,
            -40,
            feat_data['name'],
            color='#545454',
            fontsize=fontsize)

        feature_change_text = f"{feat_data['factual']}➜{feat_data['counterfactual']}"

        # Plot bar up to 50% of the plot with the factual color
        ax.add_patch(patches.PathPatch(create_bar(
            current_x, x_size), facecolor=color_bar, lw=0))

        # Plot text for feature changes
        feat_change_text = mpl.textpath.TextPath(
            (0, 0), feature_change_text, size=fontsize)
        plt.text(
            current_x + (x_size - current_x)/2 -
            feat_change_text.get_extents().width * 0.1,
            -80,
            feature_change_text,
            color='#545454',
            fontsize=fontsize)

        feat_cs_score_percentage = round(
            (feat_data['score'] - prev_score)/sum_countershapley*100, 1)
        feat_cs_score_percentage_text = mpl.textpath.TextPath(
            (0, 0), f'{feat_cs_score_percentage}%', size=fontsize)
        plt.text(
            current_x + (x_size - current_x) / 2 -
            feat_cs_score_percentage_text.get_extents().width * 0.085,
            bar_y_height_sub,
            f'{feat_cs_score_percentage}%',
            color='white',
            fontsize=fontsize,
            weight="bold")

        current_x = x_size
        prev_score = feat_data['score']

    # Print CF score
    plt.text(
        current_x,
        scale_y,
        round(features_data[-1]['score'], 2),
        color='#008ae7',
        fontsize=fontsize)

    # Draw bar for the factual score
    plt.bar(0, scale_y - 10, width=0.5, color='#ff0055', linewidth=1)
    plt.text(-5, scale_y + fontsize * 4, 'Factual Score',
             color='#ff0055', fontsize=fontsize)
    plt.text(0, scale_y, factual_score, color='#ff0055',
             fontsize=fontsize)

    # Plot bar for the counterfactual score
    plt.bar(current_x, scale_y - 10, width=0.5, color='#008ae7', linewidth=1)
    plt.text(current_x - 10, scale_y + fontsize * 4,
             'Counterfactual Score', color='#008ae7', fontsize=fontsize)

    # Remove y axis
    plt.gca().axes.get_yaxis().set_visible(False)
    plt.gca().axes.get_xaxis().set_visible(False)
    # Remove square around the plot
    plt.gca().spines['top'].set_visible(False)
    plt.gca().spines['left'].set_visible(False)
    plt.gca().spines['right'].set_visible(False)
    plt.gca().spines['bottom'].set_visible(False)

    for l_pos in x_left_pos:  # little dividers between x labels
        plt.bar(l_pos, -scale_y + 10, width=0.2, color='#cccccc', linewidth=0)
    plt.bar(scale_x, -scale_y + 10, width=0.2,
            color='#cccccc', linewidth=0)  # divider at the end

    ax.set_xlim(0, scale_x)
    ax.set_ylim(-scale_y, scale_y + 70)

    if save_path is not None:
        plt.savefig(save_path, bbox_inches='tight')

    plt.show()


def make_constellation_plot(factual_score, single_points_chart, text_features, mulitple_points_chart,
                            mulitple_points_chart_y, single_points, class_names, cf_score, point_to_pred, save_path):
    x_dim = 10
    y_dim = 4
    y_lim_low = -0.1
    y_lim_high = len(single_points_chart) - 0.9
    x_lim_low = 0
    x_lim_high = 1

    fig = plt.figure(figsize=(x_dim, y_dim))
    ax = fig.add_subplot(111)

    ax.set_xlim(x_lim_low, x_lim_high)
    ax.set_ylim(y_lim_low, y_lim_high)

    # Plot single change points
    for idx_p, p in enumerate(single_points_chart):
        plt.scatter([p[1]], [p[0]], color='#008ae7D2' if p[1]
                    >= 0.5 else '#ff0055D2', s=100)

    # Plot feature names and value changes
    max_text_features = max([len(f) for f in text_features])
    for i in range(len(text_features)):
        plt.text(-0.012*max_text_features, i,
                 text_features[i], color='#545454', fontsize=12)

    # Verify if there are multiple change points
    if len(mulitple_points_chart) > 0:
        # Plot multiple change points
        plt.scatter(mulitple_points_chart[:, 1],
                    mulitple_points_chart_y,
                    color=['#E0423A' if s <
                           0.5 else 'blue' for s in mulitple_points_chart[:, 1]],
                    s=10)

    # Plot counterfactual point
    cf_pred_x_1 = np.mean([*range(len(single_points))])
    plt.scatter([cf_score], [cf_pred_x_1], color='#008ae7D2' if cf_score
                >= 0.5 else '#ff0055D2', s=100)

    # Plot a vertical line at the threshold
    plt.axvline(x=0.5, color='#c20000', zorder=0, linewidth=0.5)

    # Plot a vertical line at the factual score
    plt.axvline(x=factual_score, color='#ff0055D2',
                linestyle='dashed', zorder=0)
    plt.text(factual_score - 0.06, len(text_features) *
             1.01 - 0.9, 'Factual Score', color='#ff0055D2')

    # Plot a vertical line at the counterfactual score
    plt.axvline(x=cf_score, color='#008ae7D2',
                linestyle='dashed', zorder=0)
    plt.text(cf_score - 0.10, len(text_features)*1.01 - 0.9,
             'Counterfactual Score', color='#008ae7D2')

    plt.gca().axes.get_yaxis().set_visible(False)
    plt.gca().spines['top'].set_visible(False)
    plt.gca().spines['left'].set_visible(False)
    plt.gca().spines['right'].set_visible(False)

    # Plot classes names
    class_0_name = class_names[0]
    class_1_name = class_names[1]
    size_factual_class = mpl.textpath.TextPath(
        (0, 0), class_0_name).get_extents().width * 0.002
    plt.text(0.48 - size_factual_class - max([len(f) for f in text_features])*0.0001,
             len(text_features) - 0.9 + len(text_features)*0.04, class_0_name, color='#ff0055D2', fontweight='bold')
    plt.text(0.49, len(text_features) - 0.9 + len(text_features)*0.04, '➜',
             color='#c20000', fontweight='bold')
    plt.text(0.51, len(text_features) - 0.9 + len(text_features)*0.04,
             class_1_name, color='#008ae7D2', fontweight='bold')

    # Plot Counterfactual lines
    for i in range(len(single_points)):
        x_0 = point_to_pred[i]
        x_1 = cf_score
        y_0 = i
        y_1 = cf_pred_x_1
        plt.plot([x_0, x_1], [y_0, y_1], color='k', zorder=0,
                 linewidth=1, alpha=0.15, linestyle='dotted')

    for points, x_value in mulitple_points_chart:
        for origin_point in points:
            x_0 = point_to_pred[origin_point]
            x_1 = x_value
            y_0 = origin_point
            y_1 = np.mean(points)

            plt.plot([x_0, x_1], [y_0, y_1], color='#e0e0e0',
                     zorder=0, linewidth=1, alpha=0.5)

    if save_path is not None:
        plt.savefig(save_path, bbox_inches='tight')



import pytest
import numpy as np
from unittest.mock import Mock

from counterplots.utils.verification import verify_factual_counterfactual_shape, verify_factual_cf, \
    verify_feature_names, verify_model_prediction, verify_binary_class, verify_class_names, verify_shape


def test_verify_shape():
    array = np.array([1, 2, 3])
    verify_shape(array)  # No exception should be raised

    array_wrong_shape = np.array([[1, 2, 3]])  # Incorrect shape
    with pytest.raises(ValueError):
        verify_shape(array_wrong_shape)


def test_verify_factual_counterfactual_shape():
    factual = np.array([1, 2, 3])
    cf = np.array([4, 5, 6])
    verify_factual_counterfactual_shape(factual, cf)  # No exception should be raised

    cf_wrong_shape = np.array([[4, 5]])  # Incorrect shape
    with pytest.raises(ValueError):
        verify_factual_counterfactual_shape(factual, cf_wrong_shape)


def test_verify_factual_cf():
    factual = np.array([1, 2, 3])
    cf = np.array([4, 5, 6])
    verify_factual_cf(factual, cf)  # No exception should be raised

    factual_wrong_values = np.array([1, 2, 3, 4])  # Different number of features
    with pytest.raises(ValueError):
        verify_factual_cf(factual_wrong_values, cf)


def test_verify_feature_names():
    factual = np.array([1, 2, 3])
    feature_names = ['f1', 'f2', 'f3']
    verify_feature_names(factual, feature_names)  # No exception should be raised

    feature_names_wrong_length = ['f1', 'f2']  # Incorrect length
    with pytest.raises(ValueError):
        verify_feature_names(factual, feature_names_wrong_length)


def test_verify_model_prediction():
    factual = np.array([1, 2, 3])
    model_pred_correct = Mock()
    model_pred_correct.predict_proba = Mock(return_value=np.array([0]))
    verify_model_prediction(model_pred_correct, factual)  # No exception should be raised

    model_pred_incorrect = ''
    with pytest.raises(ValueError):
        verify_model_prediction(model_pred_incorrect, factual)


def test_verify_binary_class():
    factual = np.array([1, 2, 3])
    model_pred = lambda x: np.array([1]) if np.sum(x) > 5 else np.array([0]) # noqa
    verify_binary_class(model_pred, factual)  # No exception should be raised

    # Multiclass prediction
    model_pred_multiclass = lambda x: np.array([[0.1, 0.5, 0.4]])  # noqa
    with pytest.raises(ValueError):
        verify_binary_class(model_pred_multiclass, factual)

    model_pred_other = lambda x: np.array([[[1]]]) # noqa
    with pytest.raises(ValueError):
        verify_binary_class(model_pred_other, factual)


def test_verify_class_names():
    class_names = {0: 'Class 0', 1: 'Class 1'}
    verify_class_names(class_names)  # No exception should be raised

    class_names_incorrect_type = 'Class Names'  # Incorrect type
    with pytest.raises(ValueError):
        verify_class_names(class_names_incorrect_type)



import numpy as np
import unittest
from unittest.mock import patch
from counterplots import CreatePlot


class CreatePlotTests(unittest.TestCase):

    def setUp(self):
        # Create dummy data for testing
        self.factual = np.array([1, 2, 3])
        self.cf = np.array([4, 5, 6])
        self.model_pred = lambda x: np.array([np.array([0]) if sum(r) < 10 else np.array([1]) for r in x])
        self.feature_names = ['feat1', 'feat2', 'feat3']
        self.class_names = {0: '0', 1: '1'}

    def test_init(self):
        create_plot = CreatePlot(self.factual, self.cf, self.model_pred, self.feature_names, self.class_names)
        self.assertTrue(np.array_equal(create_plot.factual, self.factual))
        self.assertTrue(np.array_equal(create_plot.cf, self.cf))
        self.assertEqual(create_plot.feature_names, self.feature_names)
        self.assertEqual(create_plot.class_names, self.class_names)
        self.assertEqual(create_plot.factual_score, 0)
        self.assertEqual(create_plot.cf_score, 1)

    def test_init_no_feat_name(self):
        create_plot = CreatePlot(self.factual, self.cf, self.model_pred, None, self.class_names)
        self.assertTrue(np.array_equal(create_plot.factual, self.factual))
        self.assertTrue(np.array_equal(create_plot.cf, self.cf))
        self.assertEqual(create_plot.feature_names, ['f1', 'f2', 'f3'])
        self.assertEqual(create_plot.class_names, self.class_names)
        self.assertEqual(create_plot.factual_score, 0)
        self.assertEqual(create_plot.cf_score, 1)

    def test_init_no_class_name(self):
        create_plot = CreatePlot(self.factual, self.cf, self.model_pred, self.feature_names, None)
        self.assertTrue(np.array_equal(create_plot.factual, self.factual))
        self.assertTrue(np.array_equal(create_plot.cf, self.cf))
        self.assertEqual(create_plot.feature_names, self.feature_names)
        self.assertEqual(create_plot.class_names, {0: '0', 1: '1'})
        self.assertEqual(create_plot.factual_score, 0)
        self.assertEqual(create_plot.cf_score, 1)

    @patch('counterplots.make_greedy_plot')
    @patch('counterplots.greedy_strategy')
    def test_greedy(self, mock_greedy_strategy, mock_make_greedy_plot):
        create_plot = CreatePlot(self.factual, self.cf, self.model_pred, self.feature_names, self.class_names)
        create_plot.greedy(save_path=None)  # Just check if it runs without error

    @patch('counterplots.make_countershapley_plot')
    @patch('counterplots._convert_legendvalue')
    @patch('counterplots.calc_shapley_values_between')
    def test_countershapley(self, mock_calc_shapley_values_between, mock__convert_legendvalue,
                            mock_make_countershapley_plot):

        mock_calc_shapley_values_between.return_value = ((0.4, 0.2, 0.1), (0, 1, 2))

        create_plot = CreatePlot(self.factual, self.cf, self.model_pred, self.feature_names, self.class_names)
        create_plot.countershapley(save_path=None)  # Just check if it runs without error

    @patch('counterplots.calc_shapley_values_between')
    def test_countershapley_values(self, mock_calc_shapley_values_between):
        mock_calc_shapley_values_between.return_value = ((0.4, 0.2, 0.1), (0, 1, 2))
        create_plot = CreatePlot(self.factual, self.cf, self.model_pred, self.feature_names, self.class_names)
        values = create_plot.countershapley_values()
        expected_values = {
            'feature_names': self.feature_names,
            'feature_values': (0.4, 0.2, 0.1),
            'feature_indices': (0, 1, 2)
        }
        self.assertEqual(values, expected_values)

    @patch('counterplots.make_constellation_plot')
    def test_constellation(self, mock_make_constellation_plot):
        create_plot = CreatePlot(self.factual, self.cf, self.model_pred, self.feature_names, self.class_names)
        create_plot.constellation(save_path=None)  # Just check if it runs without error

    @patch('counterplots.make_constellation_plot')
    @patch('counterplots.CreatePlot.greedy')
    def test_constellation_few_changes(self, mock_greedy, mock_make_constellation_plot):
        create_plot = CreatePlot(np.array([1, 2, 3]), np.array([1, 2, 4]),
                                 self.model_pred, self.feature_names, self.class_names)
        create_plot.constellation(save_path=None)

        # Check if greedy is called
        mock_greedy.assert_called_once_with(None)

        # Check if constellation is not called
        mock_make_constellation_plot.assert_not_called()



from itertools import combinations

import numpy as np


def create_adapted_model_and_class(factual, model_pred, class_names):
    """ Adapts the model to return a single class, factual class is also always 0

    Args:
        factual (np.ndarray): Factual instance
        model_pred (object): Model prediction function

    Returns:
        obejct: Adapted model prediction function
    """
    pred_test = np.array(model_pred(np.array([factual])))
    if len(pred_test.shape) == 2:
        # If model has two classes, return the probability of the second class (1)
        if pred_test.shape[1] == 2:
            def adapted_model(x): return np.array(model_pred(x))[:, 1] if pred_test[0][1] < 0.5 else np.array(
                model_pred(x))[:, 0]
            adapted_class = {0: class_names[0], 1: class_names[1]} if pred_test[0][1] < 0.5 else {0: class_names[1],
                                                                                                  1: class_names[0]}
        else:
            def adapted_model(x): return np.array(model_pred(x))[:, 0] if pred_test[0][0] < 0.5 else 1 - np.array(
                model_pred(x))[:, 0]
            adapted_class = {0: class_names[0], 1: class_names[1]} if pred_test[0][0] < 0.5 else {0: class_names[1],
                                                                                                  1: class_names[0]}
        return adapted_model, adapted_class

    def adapted_model(x): return np.array(model_pred(
        x)) if pred_test[0] < 0.5 else 1 - np.array(model_pred(x))
    adapted_class = {0: class_names[0], 1: class_names[1]} if pred_test[0] < 0.5 else {0: class_names[1],
                                                                                       1: class_names[0]}
    return adapted_model, adapted_class


def greedy_strategy(factual, cf, adapted_model, feature_names):
    modified_features = np.where(factual != cf)[0]

    features_data = []
    ordered_features = np.array([])
    current_factual = factual.copy()
    # Make a greedy search to find the best feature to change
    for _ in range(len(modified_features)):
        # Get features that were not added to the ordered_features
        check_features = np.array(
            [feat for feat in modified_features if feat not in ordered_features])

        # Create modified factuals with counterfactual values in the indexes of check_features
        modified_factuals = [current_factual.copy()
                             for _ in range(len(check_features))]
        for check_feature, modified_factual in zip(check_features, modified_factuals):
            modified_factual[check_feature] = cf[check_feature]

        # Make predictions
        modified_factuals_pred = adapted_model(np.array(modified_factuals))

        best_feature = check_features[np.argmax(modified_factuals_pred)]
        best_pred = np.max(modified_factuals_pred)
        best_feat_name = feature_names[best_feature]
        best_factual_feature_value = factual[best_feature]
        best_cf_feature_value = cf[best_feature]

        current_factual = modified_factuals[np.argmax(modified_factuals_pred)]

        # Adjust text of modification values
        factual_value = _convert_legendvalue(best_factual_feature_value)
        counterfactual_value = _convert_legendvalue(best_cf_feature_value)

        features_data.append({
            'score': best_pred,
            'name': best_feat_name,

            'factual': factual_value,
            'counterfactual': counterfactual_value,
        })
        ordered_features = np.append(ordered_features, best_feature)

    return features_data


def factorial(n):
    if n == 0:
        return 1
    else:
        return n * factorial(n - 1)


def calc_adapted_instance(instance, instance2, _indices):
    return [instance[i] if i not in _indices else instance2[i] for i in range(len(instance))]


def generate_contribution_graph(instance1: [], instance2: [], model, indices_feature_diff,
                                class_of_interest_index=0) -> {}:
    """
    Changes all possible combinations of features of instance1 and calculates the prediction value.
    The features to be changed are given by index by :arg indices_feature_diff: and are changed to the values as they
    appear on these positions in :arg instance2:

    :param instance1: The instance to be changed on certain features
    :param instance2: The instance whose values the features will be changed to
    :param model: The model, scoring the instance (must contain a predict_proba() method)
    :param indices_feature_diff: The indices of the features whose value in :arg instance1: need to be changed to the
    values as they appear in :arg instance2: at the same index
    :param class_of_interest_index: predict_proba() predicts an outcome for each class. Pass the index of the class of
    relevance here

    :return: Dict: dictionary whose keys are the subset of features (some combination out of :arg indices_feature_diff:)
    and whose values are the prediction value of :arg instance1:, but with adapted features on the indices of said
    subset
    """
    contribution_graph = {}
    # different lengths of subsets of features
    for L in range(0, len(indices_feature_diff) + 1):
        for subset in combinations(indices_feature_diff, L):
            changed_instance = calc_adapted_instance(
                instance1, instance2, subset)
            # you now iterate all possible feature combinations of length L
            prediction = model([changed_instance])[class_of_interest_index]
            contribution_graph[subset] = prediction
    return contribution_graph


def calc_shapley_values_between(instance1, instance2, model, class_of_interest_index=0):
    """
    Calculates shapley values between some instance and another instance, where the differences
    :param instance1: 1-dimensional array: The first instance (factual)
    :param instance2: 1-dimensional array: The second instance (counterfactual)
    :param model: the modal generating predictions. Must have a predict_proba() method
    :param class_of_interest_index: The index of the target class of interest. Generally equals 0
    :return: feature_contributions, indices
    """

    def get_shapley_feature_contribution(_shapley_graph: {}, feature_index: int):
        # amount of features/players
        n = max([len(node) for node in _shapley_graph])
        ordered_nodes = [[node for node in _shapley_graph if len(node) == i]
                         for i in range(n+1)]
        contributions_feature_i = []
        for i in range(len(ordered_nodes) - 1):
            row1 = ordered_nodes[i]
            row2 = [node for node in ordered_nodes[i + 1]
                    if feature_index in node]
            size = len(row1[0])  # |S|, i.e. the size of the set S
            for node_w_feature in row2:
                remainder = [e for e in node_w_feature if e != feature_index]
                backlink = [node for node in row1 if all(
                    [i in node for i in remainder])][0]
                # contribution of adding feature i
                marginal_contr = shapley_graph[node_w_feature] - \
                    shapley_graph[backlink]
                marginal_contr_norm = marginal_contr * \
                    factorial(size) * factorial(n-size-1) / factorial(n)
                contributions_feature_i.append(marginal_contr_norm)
        # sum all contributions over all sets S
        return np.sum(contributions_feature_i)

    # where counterfactual is different from factual
    indices_feature_diff = np.where(instance1 != instance2)[-1]
    shapley_graph = generate_contribution_graph(instance1, instance2, model, indices_feature_diff,
                                                class_of_interest_index)

    feature_contributions = []
    for feature in indices_feature_diff:
        contr = get_shapley_feature_contribution(shapley_graph, feature)
        feature_contributions.append(contr)
    feature_contributions, ind = zip(
        *sorted(zip(feature_contributions, indices_feature_diff)))
    return feature_contributions[::-1], ind[::-1]


def _format_e(x):
    return '{:.2e}'.format(x).replace('e+0', 'E').replace('e-0', 'E-').replace('e+', 'E').replace('e-', 'E-')


def _convert_legendvalue(x):
    # Verify if x is a number
    try:
        xn = float(x)
        # Verify if number has decimal places
        if xn % 1 == 0:
            # If number > 9999, convert to scientific notation
            if xn > 9999:
                return _format_e(xn)
            else:
                return x
        else:
            # If decimal places take more than 5 characters
            if abs(round(xn, 5)) <= 0.00001:
                return _format_e(xn)
            else:
                return round(xn, 5)
    except ValueError:
        return x



from setuptools import setup, find_packages

with open('README.md') as f:
    README = f.read()

with open('LICENSE.txt') as f:
    LICENSE = f.read()

setup_args = dict(
    name='counterplots',
    version='0.0.3',
    description='Plotting tool for counterfactual explanations',
    long_description_content_type='text/markdown',
    long_description=README,
    license='MIT',
    packages=find_packages(exclude=('tests\*', 'exp_notebooks\*', '_static\*')),
    author='Raphael Mazzine Barbosa de Oliveira, Bjorge Meulemeester',
    keywords=['Counterfactual Explanations', 'Visualization', 'Plotting', 'Explainable Artificial Intelligence', 'XAI', 'Machine Learning'],
    url='https://github.com/ADMAntwerp/CounterPlots',
    download_url='https://pypi.org/project/counterplots/',
)

install_requires = [
    'numpy',
    'matplotlib',
]

if __name__ == '__main__':
    setup(**setup_args, install_requires=install_requires)



import unittest
import numpy as np
from counterplots.utils.process import create_adapted_model_and_class, greedy_strategy, factorial, \
    calc_adapted_instance, generate_contribution_graph, calc_shapley_values_between, _convert_legendvalue


class TestScript(unittest.TestCase):

    def setUp(self): # noqa
        def model(data):
            out = []
            for x in data:
                if x[0] == 0.1 and x[1] == 0.2 and x[2] == 0.3:
                    out.append(0.0)
                elif x[0] == 0.4 and x[1] == 0.5 and x[2] == 0.6:
                    out.append(1.0)
                elif x[0] == 0.1 and x[1] == 0.5 and x[2] == 0.6:
                    out.append(0.4)
                elif x[0] == 0.4 and x[1] == 0.2 and x[2] == 0.6:
                    out.append(0.2)
                elif x[0] == 0.4 and x[1] == 0.5 and x[2] == 0.3:
                    out.append(0.3)
                elif x[0] == 0.1 and x[1] == 0.2 and x[2] == 0.6:
                    out.append(0.25)
                elif x[0] == 0.4 and x[1] == 0.2 and x[2] == 0.3:
                    out.append(0.4)
                elif x[0] == 0.1 and x[1] == 0.5 and x[2] == 0.3:
                    out.append(0.44)
            return np.array(out)
        self.model = model
        self.factual = np.array([0.1, 0.2, 0.3])
        self.cf = np.array([0.4, 0.5, 0.6])

    def test_create_adapted_model_and_class(self):
        class_names = ['A', 'B']
        model_pred = self.model
        factual = self.factual

        adapted_model, adapted_class = create_adapted_model_and_class(factual, model_pred, class_names)

        self.assertTrue(callable(adapted_model))
        self.assertEqual(adapted_class[0], 'A')
        self.assertEqual(adapted_class[1], 'B')

    def test_create_adapted_model_and_class_B_A(self):
        class_names = ['A', 'B']
        model_pred = self.model
        factual = self.cf

        adapted_model, adapted_class = create_adapted_model_and_class(factual, model_pred, class_names)

        self.assertTrue(callable(adapted_model))
        self.assertEqual(adapted_class[0], 'B')
        self.assertEqual(adapted_class[1], 'A')

    def test_create_adapted_model_and_class_double_pred(self):
        class_names = ['A', 'B']
        model_pred = lambda data: [[1 - self.model([x])[0], self.model([x])[0]] for x in data] # noqa
        factual = self.factual

        adapted_model, adapted_class = create_adapted_model_and_class(factual, model_pred, class_names)

        self.assertTrue(callable(adapted_model))
        self.assertEqual(adapted_class[0], 'A')
        self.assertEqual(adapted_class[1], 'B')

    def test_create_adapted_model_and_class_double_pred_B_A(self):
        class_names = ['A', 'B']
        model_pred = lambda data: [[1 - self.model([x])[0], self.model([x])[0]] for x in data] # noqa
        factual = self.cf

        adapted_model, adapted_class = create_adapted_model_and_class(factual, model_pred, class_names)

        self.assertTrue(callable(adapted_model))
        self.assertEqual(adapted_class[0], 'B')
        self.assertEqual(adapted_class[1], 'A')

    def test_greedy_strategy(self):
        factual = self.factual
        cf = self.cf
        adapted_model = self.model
        feature_names = ['X', 'Y', 'Z']

        features_data = greedy_strategy(factual, cf, adapted_model, feature_names)

        self.assertEqual(features_data,
                         [{'score': 0.44, 'name': 'Y', 'factual': 0.2, 'counterfactual': 0.5},
                          {'score': 0.4, 'name': 'Z', 'factual': 0.3, 'counterfactual': 0.6},
                          {'score': 1.0, 'name': 'X', 'factual': 0.1, 'counterfactual': 0.4}])

    def test_factorial(self):
        self.assertEqual(factorial(0), 1)
        self.assertEqual(factorial(1), 1)
        self.assertEqual(factorial(5), 120)

    def test_calc_adapted_instance(self):
        instance = self.factual
        instance2 = self.cf
        indices = [0, 2]

        adapted_instance = calc_adapted_instance(instance, instance2, indices)

        self.assertEqual(adapted_instance, [0.4, 0.2, 0.6])

    def test_generate_contribution_graph(self):
        instance1 = self.factual
        instance2 = self.cf
        model = lambda x: np.array([0.7]) # noqa
        indices_feature_diff = [0, 2]
        class_of_interest_index = 0

        contribution_graph = generate_contribution_graph(instance1, instance2, model, indices_feature_diff,
                                                         class_of_interest_index)

        self.assertEqual(len(contribution_graph), 4)
        self.assertIn((), contribution_graph)
        self.assertIn((0,), contribution_graph)
        self.assertIn((2,), contribution_graph)
        self.assertIn((0, 2), contribution_graph)

    def test_calc_shapley_values_between(self):
        instance1 = self.factual
        instance2 = self.cf
        model = self.model

        class_of_interest_index = 0

        feature_contributions, indices = calc_shapley_values_between(
            instance1, instance2, model, class_of_interest_index)

        self.assertEqual(feature_contributions, (0.42166666666666663, 0.30166666666666664, 0.2766666666666666))
        self.assertEqual(indices, (1, 0, 2))

    def test__convert_legendvalue(self):
        self.assertEqual(_convert_legendvalue(10000), '1.00E4')
        self.assertEqual(_convert_legendvalue(0.00001), '1.00E-5')
        self.assertEqual(_convert_legendvalue(0.123456), 0.12346)
        self.assertEqual(_convert_legendvalue('ABC'), 'ABC')



from itertools import combinations
from typing import Callable, List, Dict

import numpy as np

from counterplots.utils.plots import make_greedy_plot, make_countershapley_plot, make_constellation_plot
from counterplots.utils.process import create_adapted_model_and_class, greedy_strategy, calc_shapley_values_between, \
    _convert_legendvalue
from counterplots.utils.verification import verify_factual_counterfactual_shape, verify_factual_cf,  \
    verify_feature_names, verify_model_prediction, verify_binary_class, verify_class_names


class CreatePlot:

    def __init__(
            self,
            factual: np.ndarray,
            cf: np.ndarray,
            model_pred: Callable,
            feature_names: List[str] = None,
            class_names: Dict[int, str] = None):

        # If feature_names is not provided, create it
        if feature_names is None:
            feature_names = [f'f{i+1}' for i in range(len(factual))]

        # If class_names is not provided, create it
        if class_names is None:
            class_names = {0: '0', 1: '1'}

        # Convert factual and cf to numpy arrays
        self.factual = np.array(factual)
        self.cf = np.array(cf)
        self.feature_names = feature_names
        self.class_names = class_names

        # Make verifications
        verify_factual_counterfactual_shape(factual, cf)
        verify_factual_cf(factual, cf)
        verify_feature_names(factual, feature_names)
        verify_model_prediction(model_pred, factual)
        verify_binary_class(model_pred, factual)
        verify_class_names(class_names)

        # Adapt model and class names
        self.adapted_model, self.class_names = create_adapted_model_and_class(
            factual=factual,
            model_pred=model_pred,
            class_names=class_names)

        self.factual_score = round(self.adapted_model(
            self.factual.reshape(1, -1))[0], 2)
        self.cf_score = round(self.adapted_model(self.cf.reshape(1, -1))[0], 2)

    def greedy(self, save_path: str = None):
        # Make a greedy search to find the best feature to change
        features_data = greedy_strategy(
            factual=self.factual,
            cf=self.cf,
            adapted_model=self.adapted_model,
            feature_names=self.feature_names)

        # Plot the greedy search
        make_greedy_plot(
            factual_score=self.factual_score,
            features_data=features_data,
            class_names=self.class_names,
            save_path=save_path)

    def countershapley(self, save_path: str = None):

        feat_contributions, feat_modified = calc_shapley_values_between(
            instance1=self.factual,
            instance2=self.cf,
            model=self.adapted_model,
            class_of_interest_index=0)

        # Create data for countershapley plot
        sum_contributions = np.sum(feat_contributions)
        current_score = self.factual_score
        features_data = []
        for feat, contrib in zip(feat_modified, feat_contributions):
            current_score += contrib

            factual_raw_value = self.factual[feat]
            cf_value = self.cf[feat]

            # Adjust text of modification values
            factual_value = _convert_legendvalue(factual_raw_value)
            counterfactual_value = _convert_legendvalue(cf_value)

            features_data.append({
                'x': contrib/sum_contributions*100,
                'score': current_score,
                'name': self.feature_names[feat],

                'factual': factual_value,

                'counterfactual': counterfactual_value})

        # Plot the countershapley values
        make_countershapley_plot(
            self.factual_score, features_data, self.class_names, save_path)

    def countershapley_values(self):
        feat_contributions, feat_modified = calc_shapley_values_between(
            instance1=self.factual,
            instance2=self.cf,
            model=self.adapted_model,
            class_of_interest_index=0)

        return {
            'feature_names': [self.feature_names[feat] for feat in feat_modified],
            'feature_values': feat_contributions,
            'feature_indices': feat_modified
        }

    def constellation(self, save_path: str = None):

        # Get the index of the features that are different between the factual and counterfactual
        idx_feat_diff = np.where(self.factual != self.cf)[0]

        # Constellation need at least 2 features to be different
        if len(idx_feat_diff) < 2:
            print(
                'Showing greedy plot since constellation needs at least 2 features to be different')
            self.greedy(save_path)
            return

        text_features = []
        for i in idx_feat_diff:

            # Adjust text of modification values
            factual_raw_value = self.factual[i]
            cf_value = self.cf[i]

            factual_value = _convert_legendvalue(factual_raw_value)
            counterfactual_value = _convert_legendvalue(cf_value)

            factual_feat_text = str(factual_value)
            cf_feat_text = str(counterfactual_value)
            text_features.append(
                f'{self.feature_names[i]} ({factual_feat_text}➜{cf_feat_text})')

        # Create all possible combinations from 1 to the number of features that are different - 1
        idx_feat_comb = [list(combinations(idx_feat_diff, i))
                         for i in range(1, len(idx_feat_diff))]
        idx_feat_comb = [item for sublist in idx_feat_comb for item in sublist]

        # Dataset with modified points
        points_mod = np.array([self.factual] * len(idx_feat_comb))

        # Make the changes to the points
        points_mod = np.array(
            [np.array([points_mod[i][j] if j not in idx_feat_comb[i] else self.cf[j]
                       for j in range(len(self.factual))]) for i in range(len(idx_feat_comb))])

        # Get points predictions
        points_pred = self.adapted_model(points_mod)

        # Get single points with single change and their predictions
        single_points = np.array(
            [[comb[0], points_pred[idx]] for idx, comb in enumerate(idx_feat_comb) if len(comb) == 1])

        # Get multiple points with multiple changes and their predictions
        multiple_points = np.array(
            [[comb, points_pred[idx]] for idx, comb in enumerate(idx_feat_comb) if len(comb) > 1], dtype=object)

        # Points to chart points
        points_to_chart = {p: i for i, p in enumerate(single_points[:, 0])}
        point_to_pred = {i: p for i, p in enumerate(single_points[:, 1])}

        # Convert point features to chart points
        single_points_chart = np.array(
            [[points_to_chart[point_feat], point_pred] for point_feat, point_pred in single_points])
        mulitple_points_chart = np.array(
            [[np.array([points_to_chart[pf] for pf in point_feat]), point_pred] for point_feat, point_pred in
             multiple_points], dtype=object)

        # Multiple points y axis (0 index of mulitple_points_chart) must be the mean of the point values
        mulitple_points_chart_y = [np.mean(fp)
                                   for fp, _ in mulitple_points_chart]

        make_constellation_plot(
            factual_score=self.factual_score,
            single_points_chart=single_points_chart,
            text_features=text_features,
            mulitple_points_chart=mulitple_points_chart,
            mulitple_points_chart_y=mulitple_points_chart_y,
            single_points=single_points,
            class_names=self.class_names,
            cf_score=self.cf_score,
            point_to_pred=point_to_pred,
            save_path=save_path)
